// BLE Nano (V1.5) Remote Light Switch Control
// 19th September 2017 -- version L1
/*
   (c)2014-2017 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#include <BLE_API.h>
// download the pfodParser library V3.10+ from http://www.forward.com.au/pfod/pfodParserLibraries/index.html
#include <pfodParser.h>
#include "PushButton.h"

// DEVICE_NAME 20 char or less
#define DEVICE_NAME            "Garage Light"

// this enables Serial debug
// It uses about 1mA more current, but makes the connection more reliable (why?)
// comment this out if you wish but you may get some drop outs and re-connects.
// Average current with DEBUG ~1.5mA
// Note: set pfodApp connection timeout to 3sec so any lack of response is detected quickly and pfodApp reconects promptly
//#define DEBUG

pfodParser parser("L1"); // create a parser to handle the pfod messages
pfodDwgs dwgs(&parser);  // drawing support
PushButton pushButton('a', &dwgs);

// 1.5mA drops down to 0.7mA
// adding Serial.begin() adds about 0.7mA  i.e. stays around 1.5mA
// give the board pins names, if you change the pin number here you will change the pin controlled
int cmd_C_var; // name the variable for 'Reading'
int cmd_OnOff_var; // onOff cmd
bool triggerD6 = false; // if true pulse D6 else if false pulse D7

const int cmd_C_pin = A5; // name A5 for 'Reading'

unsigned long cmd_B_PULSE_LENGTH = 20; // 0.020 secs
const int cmd_B_pin = D5; // name the output D5 for 'Set Pulse' //D6 does NOT work goes high on startup and kills volts due to trying to drive relay
unsigned long cmd_D_PULSE_LENGTH = 20; // 0.020 secs
const int cmd_D_pin = D7; // name the output D7 for 'Reset Pulse'

#define TXRX_BUF_LEN           20

BLE                            ble;
Timeout                        timeout;
Timeout  pulseTimer; // for set/reset pulses only one running at a time
Timeout  blockingTimer; // blocks double switches
static bool blocking = false; // not blocking
const unsigned long BLOCKING_PULSE_LENGTH = 300; // 0.30 secs

Timeout  offTimer; // turn off light after timeout
static int offTimerCounter = 0; // timer disabled or has timed out
const unsigned long OFF_TIMER_COUNTER_LENGTH = 60; // 60 secs per count
// Set OFF_TIMEOUT_IN_MINS to 0 to disable auto turn off
const int OFF_TIMEOUT_IN_MINS = 10; // turn off light in 10 mins

Timeout  switchDelayTimer; // delays relay pulse until replay BLE msg sent
static bool delaying = false; // not in delay
const unsigned long SWITCH_DELAY_LENGTH = 250; // 0.250 secs NOTE: less than BLOCKING_PULSE_LENGTH.

static uint8_t rx_buf[TXRX_BUF_LEN];
static uint8_t rx_buf_num;
static uint8_t rx_state = 0;

uint8_t tx_value[TXRX_BUF_LEN] = {0,};
uint8_t rx_value[TXRX_BUF_LEN] = {0,};


uint8_t fullTxBuffer[1024] = {0,};
size_t fullTxIdx = 0;
size_t TxIdx = 0; // goes from 0 to fullTxIdx;

void updatePushButtonOnOffState();

// Initialize value of chars
GattCharacteristic  characteristicTX(UARTServiceTXCharacteristicUUID, tx_value, 1, TXRX_BUF_LEN, GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_WRITE | GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_WRITE_WITHOUT_RESPONSE );
GattCharacteristic  characteristicRX(UARTServiceRXCharacteristicUUID, rx_value, 1, TXRX_BUF_LEN, GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_NOTIFY);
GattCharacteristic *uartChars[] = {&characteristicTX, &characteristicRX};
GattService         uartService(UARTServiceUUID, uartChars, sizeof(uartChars) / sizeof(GattCharacteristic *));

// =========== pfodBLESerial definitions
const char* localName = DEVICE_NAME;  // <<<<<<  change this string to customize the adverised name of your board (max 20 chars)
class pfodBLESerial : public Stream {
  public:
    pfodBLESerial();
    void begin(); void poll(); size_t write(uint8_t); size_t write(const uint8_t*, size_t); int read();
    int available(); void flush(); int peek(); void close(); bool isConnected();
    static void connectHandler();
    static void disconnectHandler();
    static void receiveHandler(uint8_t* buf, uint16_t bytesRead);
    static const int BLE_RX_MAX_LENGTH = 256;
    static size_t txIdx;
    static const int BLE_MAX_LENGTH = 20;
    static uint8_t txBuffer[BLE_MAX_LENGTH];
  private:
    const static uint8_t pfodEOF[1]; const static char* pfodCloseConnection;
    static volatile size_t rxHead;
    static volatile size_t rxTail;
    volatile static uint8_t rxBuffer[BLE_RX_MAX_LENGTH];
    volatile static bool connected;
    static void addReceiveBytes(const uint8_t* bytes, size_t len);
};

volatile size_t pfodBLESerial::rxHead = 0; volatile size_t pfodBLESerial::rxTail = 0;
volatile uint8_t pfodBLESerial::rxBuffer[BLE_RX_MAX_LENGTH];
size_t pfodBLESerial::txIdx = 0;
uint8_t pfodBLESerial::txBuffer[pfodBLESerial::BLE_MAX_LENGTH];
const uint8_t pfodBLESerial::pfodEOF[1] = {(uint8_t) - 1};
const char* pfodBLESerial::pfodCloseConnection = "{!}"; volatile bool pfodBLESerial::connected = false;

pfodBLESerial bleSerial; // create a BLE serial connection


void disconnectionCallBack(const Gap::DisconnectionCallbackParams_t *params) {
#ifdef DEBUG
  Serial.println("Disconnected!");
#endif
  pfodBLESerial::disconnectHandler();
#ifdef DEBUG
  Serial.println("Restarting the advertising process");
#endif
  ble.startAdvertising();
}

void connectionCallBack(const Gap::ConnectionCallbackParams_t *params) {
#ifdef DEBUG
  Serial.println("Conncted!");
#endif
  pfodBLESerial::connectHandler();
}


void gattServerWriteCallBack(const GattWriteCallbackParams *Handler) {
  uint8_t buf[TXRX_BUF_LEN];
  uint16_t bytesRead, index;

#ifdef DEBUG
  // Serial.println("onDataWritten : ");
#endif
  if (Handler->handle == characteristicTX.getValueAttribute().getHandle()) {
    ble.readCharacteristicValue(characteristicTX.getValueAttribute().getHandle(), buf, &bytesRead);
#ifdef DEBUG
    //    Serial.print("bytesRead: ");
    //    Serial.println(bytesRead, HEX);
    for (index = 0; index < bytesRead; index++) {
      Serial.write(buf[index]);
    }
    Serial.println("");
#endif
  }
  pfodBLESerial::receiveHandler(buf, bytesRead);
}

volatile int attached = 0; // set when sendtimer attached

void timeOut_handle() {   //update characteristic data
#ifdef DEBUG
  //  Serial.print("O:");
  //  Serial.println(millis());
#endif
  attached = 0;
  sendNext20Bytes();
}

void checkTurnOff() {
#ifdef DEBUG
  Serial.print("called checkTurnOff Counter:");
  Serial.println(offTimerCounter);
#endif
  updatePushButtonOnOffState(); // check if light on or off
  if (!pushButton.isOn()) {
    offTimerCounter = 0; // stop
  } else {
    // light is ON
    if (offTimerCounter <= 0) {
      // not counting but should be
      offTimerCounter = OFF_TIMEOUT_IN_MINS + 1; // start timer
    }
    // else counting down
  }

  offTimer.attach(checkTurnOff, OFF_TIMER_COUNTER_LENGTH); // start next timer to fire in 1min
  if (offTimerCounter <= 0) {
    return; // not counting
  }
  // else
  if (offTimerCounter == 1) {
    // turn light off
#ifdef DEBUG
    Serial.println(" turn light off");
#endif
    turnLightOff();
  }
  offTimerCounter--; // decrement  if goes to 0 then finished
}

void turnLightOff() {
  if (pushButton.isOn()) { // light is ON
    // toggel to swtich off;
    pushButton.setOnState(!pushButton.isOn()); // set state for dwg update
    triggerD6 = !triggerD6;
    switchRelay();
  }
}

void writeRXbuffer() {
  ble.updateCharacteristicValue(characteristicRX.getValueAttribute().getHandle(), pfodBLESerial::txBuffer, pfodBLESerial::txIdx);
  memset(pfodBLESerial::txBuffer, 0x00, pfodBLESerial::BLE_MAX_LENGTH);
  // rx_state = 0;
}

void setup() {

  // put your setup code here, to run once
  // initialize io here because if set RTS after Serial.begin then system reboots!! ??  and this set is run twice.
  cmd_C_var = 0;
  pinMode(cmd_B_pin, OUTPUT); // output for 'Set Pulse' is initially LOW,
  digitalWrite(cmd_B_pin, 0); // set output
  pinMode(cmd_D_pin, OUTPUT); // output for 'Reset Pulse' is initially LOW,
  digitalWrite(cmd_D_pin, 0); // set output

  //  for (int i=0; i<3; i++) {
  //    delay(1000);
  //  }

#ifdef DEBUG
  Serial.begin(9600);  // this adds about 0.7mA but improves connection reliability (why?)
#endif

  ble.init();
  ble.onDisconnection(disconnectionCallBack);
  ble.onConnection(connectionCallBack);
  ble.onDataWritten(gattServerWriteCallBack);
  //  ble.onDataRead(readCallBack);

  // setup adv_data and srp_data
  ble.accumulateAdvertisingPayload(GapAdvertisingData::BREDR_NOT_SUPPORTED);
  ble.accumulateAdvertisingPayload(GapAdvertisingData::SHORTENED_LOCAL_NAME,
                                   (const uint8_t *)DEVICE_NAME, sizeof(DEVICE_NAME) - 1);
  ble.accumulateAdvertisingPayload(GapAdvertisingData::COMPLETE_LIST_128BIT_SERVICE_IDS,
                                   (const uint8_t *)UARTServiceUUID_reversed, sizeof(UARTServiceUUID_reversed));
  // set adv_type
  ble.setAdvertisingType(GapAdvertisingParams::ADV_CONNECTABLE_UNDIRECTED);
  // add service
  ble.addService(uartService);
  // set device name
  ble.setDeviceName((const uint8_t *)DEVICE_NAME);
  // set tx power,valid values are -40, -20, -16, -12, -8, -4, 0, 4
  ble.setTxPower(4);
  // set adv_interval, 100ms in multiples of 0.625ms.
  ble.setAdvertisingInterval(160);
  // set adv_timeout, in seconds
  ble.setAdvertisingTimeout(0);
  // start advertising
  ble.startAdvertising();
#ifdef DEBUG
  //  Serial.println("Advertising Started!");
#endif


  bleSerial.begin();
  parser.connect(&bleSerial);
#ifdef DEBUG
  Serial.println("Setup Complete.");
#else
  // use this block if not using Serial
  NRF_POWER->DCDCEN = 0x00000001;
  NRF_UART0->TASKS_STOPTX = 1;
  NRF_UART0->TASKS_STOPRX = 1;
  NRF_UART0->ENABLE = 0;
#endif

  offTimer.attach(checkTurnOff, OFF_TIMER_COUNTER_LENGTH); // start turn off timer fire in 1min
}

void sendNext20Bytes() {
  if (TxIdx < fullTxIdx) {
    pfodBLESerial::txIdx = 0;
    while ((TxIdx < fullTxIdx) && (pfodBLESerial::txIdx < sizeof(pfodBLESerial::txBuffer))) {
      pfodBLESerial::txBuffer[pfodBLESerial::txIdx++] = fullTxBuffer[TxIdx++];
    }
    writeRXbuffer();
    if (TxIdx < fullTxIdx) {
      if (attached == 0) {
        attached = 1;
        timeout.attach_us(timeOut_handle, 200000); // send next one in 0.2 secs
      }
    } else {
      TxIdx = 0; fullTxIdx = 0;
    }
  } else {
    TxIdx = 0; fullTxIdx = 0; // fail safe
  }
}

void updatePushButtonOnOffState() {
  cmd_C_readADC();
  if (cmd_C_var < 124) { // 124/1024 * 3.3V == 400mV, measured 0.48V at A5 test point
    // the power is ON
    pushButton.setOnState(true);
  } else {
    pushButton.setOnState(false);
  }
  //#ifdef DEBUG
  //  Serial.println("Set lightOn for testing");
  //#endif
  //  pushButton.setOnState(true); // testing
}

void loop() {
  ble.waitForEvent(); // blocks (sleeps) until something happens, get msg for example

  updatePushButtonOnOffState();

  byte cmd = parser.parse(); // parse incoming data from connection
  // parser returns non-zero when a pfod command is fully parsed
  if (cmd != 0) { // have parsed a complete msg { to }
#ifdef DEBUG
    //   Serial.println((char)cmd);
#endif
    byte* pfodFirstArg = parser.getFirstArg(); // may point to \0 if no arguments in this msg.
    long pfodLongRtn; // used for parsing long return arguments, if any
    if ('.' == cmd) {
      // pfodApp has connected and sent {.} , it is asking for the main menu
      if (!parser.isRefresh()) {
        sendMainMenu(); // send back the menu designed
      } else {
        sendMainMenuUpdate(); // menu is cached just send update
      }

    } else if ('A' == cmd) {
      // in the main Menu of Menu_1
      char dwgCmd = parser.parseDwgCmd(); // parse rest of dwgCmd, return first char of active cmd
      if (pushButton.getCmd() == dwgCmd) { // user touched LED control
        if (!blocking) { // OK do this
          // sometimes BLE serves up the same msg twice(why?).
          // seems to happen more often on turning ligh off.  This block prevents the processing of the second msg turning the light on again
          // may not be needed now that the relay switch is being delayed slightly
          blocking = true;
          blockingTimer.attach_us(blockingTimeout_check, BLOCKING_PULSE_LENGTH * 1000); // turn off blocking in 0.3S
          pushButton.setOnState(!pushButton.isOn()); // set state for dwg update
          triggerD6 = !triggerD6;
          switchDelayTimer.attach_us(switchRelay, SWITCH_DELAY_LENGTH * 1000); // delay relay pulse for 0.25s to let BLE respone send in the clear
          sendDrawingUpdates_z(); // update the drawing
        } else {
          parser.print(F("{}"));
        }
      } else {
        parser.print(F("{}"));
      }
      // always send back a response or pfodApp will timeout
      // if the same msg was processed twice, pfodApp will just drop this extra return.


    } else if ('z' == cmd) { // this is the dwg in menu item 'A'
      // get starting offset
      if (!parser.isRefresh()) {
        sendDrawing_z();
      } else {
        sendDrawingUpdates_z();
      }

    } else if ('!' == cmd) {
      // CloseConnection command
      closeConnection(parser.getPfodAppStream());
    } else {
      // unknown command
      parser.print(F("{}")); // always send back a pfod msg otherwise pfodApp will disconnect.
    }
    sendNext20Bytes(); //send first part of response NOW
  }

  if ((TxIdx < fullTxIdx) && (TxIdx == 0) && (attached == 0)) {
    attached = 1;
    timeout.attach_us(timeOut_handle, 200000); // send next one in 0.2 secs
  }
}

void switchRelay() {
#ifdef DEBUG
  // Serial.println("Start Relay Switch");
#endif
  if (triggerD6) {
    digitalWrite(cmd_B_pin, 1); // set output D5
    pulseTimer.attach_us(cmd_B_checkPulse, cmd_B_PULSE_LENGTH * 1000); // turn pulse off in 20mS

  } else {
    digitalWrite(cmd_D_pin, 1); // set output D7
    pulseTimer.attach_us(cmd_D_checkPulse, cmd_D_PULSE_LENGTH * 1000); // turn pulse off in 25mS
  }
}

void cmd_C_readADC() {
  cmd_C_var = analogRead(cmd_C_pin);  // read ADC input
}

void blockingTimeout_check() {
  blocking = false; // cancel blocking
#ifdef DEBUG
  //  Serial.println("Blocking Cancelled");
#endif

}

void cmd_B_checkPulse() {
  digitalWrite(cmd_B_pin, 0); // update output pin
}

void cmd_D_checkPulse() {
  digitalWrite(cmd_D_pin, 0); // update output pin
}

void sendMainMenu() {
  // !! Remember to change the parser version string
  //    every time you edit this method
  parser.print(F("{,"));  // start a Menu screen pfod message
  // send menu background, format, prompt, refresh and version
  parser.print(F("<bg n><b><+4>~"));
  parser.print(F(DEVICE_NAME));
  parser.print(F("`5000"));
  parser.sendVersion(); // send the menu version
  // send menu items
  parser.print(F("|+A~z")); // menu item 'A' loads dwg 'z'
  parser.print(F("}"));  // close pfod message
}

void sendMainMenuUpdate() {
  parser.print(F("{;"));  // start an Update Menu pfod message
  // send menu items
  parser.print(F("|A"));
  parser.print(F("}"));  // close pfod message
  // ============ end of menu ===========
}

void sendDrawing_z() {
  dwgs.start(50, 60, dwgs.NAVY); // background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendVersion(); // send the parser version to cache this image
  dwgs.pushZero(25, 30, 3.0f); // scale by 1.0f,  scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  pushButton.draw();  // button drawn
  dwgs.popZero();
  dwgs.end();
}

void sendDrawingUpdates_z() {
  dwgs.startUpdate();
  pushButton.update();  // update existing button with current state
  dwgs.end();
}


void closeConnection(Stream *io) {
  // add any special code here to force connection to be dropped
  ((pfodBLESerial*)io)->close();
}

// ========== pfodBLESerial methods
pfodBLESerial::pfodBLESerial()  {
};

bool pfodBLESerial::isConnected() {
  return (connected);// && txCharacteristic.subscribed());
}
void pfodBLESerial::begin() {
}

void pfodBLESerial::close() {
  ble.disconnect(Gap::LOCAL_HOST_TERMINATED_CONNECTION);
#ifdef DEBUG
  Serial.println("called ble.disconnect()");
#endif
}

size_t pfodBLESerial::write(const uint8_t* bytes, size_t len) {
  for (size_t i = 0; i < len; i++) {
    write(bytes[i]);
  }
  return len; // just assume it is all written
}

size_t pfodBLESerial::write(uint8_t b) {
  if (!isConnected()) {
    return 1;
  }
  if (fullTxIdx >= sizeof(fullTxBuffer)) {
    return 1;
  }
#ifdef DEBUG
  // Serial.print((char)b);
#endif

  fullTxBuffer[fullTxIdx++] = b;
  //  if ((fullTxIdx == sizeof(fullTxBuffer)) || (b == ((uint8_t)'\n')) || (b == ((uint8_t)'}')) ) {
  //    //  flush(); // send this buffer if full or end of msg or rawdata newline
  //  }
  //  // txBuffer[txIdx++] = b;
  //  //  if ((txIdx == sizeof(txBuffer)) || (b == ((uint8_t)'\n')) || (b == ((uint8_t)'}')) ) {
  //  //    flush(); // send this buffer if full or end of msg or rawdata newline
  //  //  }
  return 1;
}

int pfodBLESerial::read() {
  if (rxTail == rxHead) {
    return -1;
  }
  // note increment rxHead befor writing
  // so need to increment rxTail befor reading
  rxTail = (rxTail + 1) % sizeof(rxBuffer);
  uint8_t b = rxBuffer[rxTail];
  return b;
}

// called as part of parser.parse() so will poll() each loop()
int pfodBLESerial::available() {
  //  flush(); // send any pending data now. This happens at the top of each loop()
  int rtn = ((rxHead + sizeof(rxBuffer)) - rxTail ) % sizeof(rxBuffer);
  return rtn;
}

void pfodBLESerial::flush() {
}

int pfodBLESerial::peek() {
  // BLEPeripheral::poll();
  if (rxTail == rxHead) {
    return -1;
  }
  size_t nextIdx = (rxTail + 1) % sizeof(rxBuffer);
  uint8_t byte = rxBuffer[nextIdx];
  return byte;
}

void pfodBLESerial::connectHandler() {
  // clear parser with -1 incase partial message left
  // should not be one
  TxIdx = 0; fullTxIdx = 0;
  addReceiveBytes(pfodEOF, sizeof(pfodEOF));
  connected = true;
  attached = 0; // set when sendtimer attached

}

void pfodBLESerial::disconnectHandler() {
  // parser.closeConnection();
  // clear parser with -1 and insert {!} incase connection just lost
  addReceiveBytes(pfodEOF, sizeof(pfodEOF));
  addReceiveBytes((const uint8_t*)pfodCloseConnection, sizeof(pfodCloseConnection));
  TxIdx = 0; fullTxIdx = 0;
  connected = false;
  attached = 0; // set when sendtimer attached

}

void pfodBLESerial::addReceiveBytes(const uint8_t* bytes, size_t len) {
  // note increment rxHead befor writing
  // so need to increment rxTail befor reading
  for (size_t i = 0; i < len; i++) {
    rxHead = (rxHead + 1) % sizeof(rxBuffer);
    rxBuffer[rxHead] = bytes[i];
  }
}


void pfodBLESerial::receiveHandler(uint8_t* buf, uint16_t bytesRead) {
  size_t len = bytesRead;
  addReceiveBytes((const uint8_t*)buf, len);
}
//======================= end pfodBLESerial methods


